// -*- c++ -*-
#ifndef __HASH_MAP_H
#define __HASH_MAP_H

#include <stdlib.h>

template<class A>
inline void swap(A& a, A& b)
{
  A c = a;
  a = b;
  b = c;
}

inline int hash(int h)
{
  h ^= h >> 16;
  h *= 0x85ebca6b;
  h ^= h >> 13;
  h *= 0xc2b2ae35;
  h ^= h >> 16;
  return h;
}

inline int upper_power_of_two(int v)
{
  v--;
  v |= v >> 1;
  v |= v >> 2;
  v |= v >> 4;
  v |= v >> 8;
  v |= v >> 16;
  v++;
  return v;
}

#define FIRST_CELL(h) (cells + ((h) & (arraySize - 1)))
#define CIRCULAR_NEXT(c) ((c) + 1 != cells + arraySize ? (c) + 1 : cells)
#define CIRCULAR_OFFSET(a, b) ((b) >= (a) ? (b) - (a) : arraySize + (b) - (a))

template<class Key, class Value> class HashMap
{
  HashMap(const HashMap<Key, Value>&);
  HashMap<Key, Value>& operator=(const HashMap<Key, Value>&);

public:
  struct Cell {
    Cell() : key() {}
    Key key;
    Value value;
  };

private:
  Cell* cells;
  int arraySize;
  int population;
  bool zeroUsed;
  Cell zeroCell;

  void repopulate(int desiredSize)
  {
    if (desiredSize == 0) {
      desiredSize = 8;
    }
    Cell* oldCells = cells;
    Cell* end = cells + arraySize;

    arraySize = desiredSize;
    cells = new Cell[arraySize];
    for (Cell* c = oldCells; c != end; c++) {
      if (c->key != zeroCell.key) {
        for (Cell* cell = FIRST_CELL(hash(c->key));; cell = CIRCULAR_NEXT(cell)) {
          if (cell->key == zeroCell.key) {
            *cell = *c;
            break;
          }
        }
      }
    }
    delete [] oldCells;
  }

public:
  HashMap() : cells(0), arraySize(0), population(0), zeroUsed(false)
  {
  }

  HashMap(HashMap<Key, Value>&& a)
    : arraySize(a.arraySize), population(a.population), zeroUsed(a.zeroUsed),
      zeroCell(a.zeroCell), cells(a.cells)
  {
    a.arraySize = a.population = 0;
    a.zeroUsed = false;
    a.zeroCell = Cell();
    a.cells = 0;
  }

  HashMap<Key, Value>& operator=(HashMap<Key, Value>&& a)
  {
    swap(arraySize, a.arraySize);
    swap(population, a.population);
    swap(zeroUsed, a.zeroUsed);
    swap(zeroCell, a.zeroCell);
    swap(cells, a.cells);
    return *this;
  }

  ~HashMap() { delete [] cells; }

  bool empty() const
  {
    return population == 0;
  }

  int size() const
  {
    return population;
  }

  const Value* find(const Key& key) const
  {
    if (key != zeroCell.key && cells) {
      for (Cell* cell = FIRST_CELL(hash(key));; cell = CIRCULAR_NEXT(cell)) {
        if (cell->key == key) {
          return &cell->value;
        }
        if (cell->key == zeroCell.key) {
          return 0;
        }
      }
    } else if (zeroUsed) {
      return &zeroCell.value;
    }
    return 0;
  }

  const Value& operator[](const Key& key) const
  {
    const Value* v = find(key);
    return *v;
  }

  Value* find(const Key& key)
  {
    if (key != zeroCell.key && cells) {
      for (Cell* cell = FIRST_CELL(hash(key));; cell = CIRCULAR_NEXT(cell)) {
        if (cell->key == key) {
          return &cell->value;
        }
        if (cell->key == zeroCell.key) {
          return 0;
        }
      }
    } else if (zeroUsed) {
      return &zeroCell.value;
    }
    return 0;
  }

  Value& operator[](const Key& key)
  {
    Value* v = find(key);
    return *v;
  }

  void insert(const Key& key, const Value& value)
  {
    if (key != zeroCell.key) {
      if (!cells) {
        repopulate(0);
      }
      for (;;) {
        for (Cell* cell = FIRST_CELL(hash(key));; cell = CIRCULAR_NEXT(cell)) {
          if (cell->key == key) {
            //abort();      // vkladat lze jen klicen ktere jeste v mape nejsou!
            cell->value = value;
            return;
          }
          if (cell->key == zeroCell.key) {
            if ((population + 1) * 4 >= arraySize * 3) {
              repopulate(arraySize * 2);
              break;
            }
            ++population;
            cell->key = key;
            cell->value = value;
            return;
          }
        }
      }
    } else {
      if (!zeroUsed) {
        zeroUsed = true;
        if (++population * 4 >= arraySize * 3) {
          repopulate(arraySize * 2);
        }
      }
      zeroCell.value = value;
    }
  }

  void clear()      // podobne jako Array::clear()
  {
    for (int i = 0; i < arraySize; i++) {
      cells[i].key = zeroCell.key;
      cells[i].value = Value();
    }
    population = 0;
    zeroCell.value = Value();
    zeroUsed = false;
  }

  void freemem()     // uvolni vsechnu pamet
  {
    delete [] cells;
    cells = 0;
    arraySize = population = 0;
    zeroCell.value = Value();
    zeroUsed = false;
  }

  void compact()
  {
    repopulate(upper_power_of_two((population * 4 + 3) / 3));
  }

  void remove(const Key& key)
  {
    if (!cells) {
      return;
    }
    if (key != zeroCell.key) {
      for (Cell* cell = FIRST_CELL(hash(key));; cell = CIRCULAR_NEXT(cell)) {
        if (cell->key == key) {
          for (Cell* neighbor = CIRCULAR_NEXT(cell);;
               neighbor = CIRCULAR_NEXT(neighbor)) {
            if (neighbor->key == zeroCell.key) {
              cell->key = zeroCell.key;
              cell->value = Value();
              population--;
              return;
            }
            Cell* ideal = FIRST_CELL(hash(neighbor->key));
            if (CIRCULAR_OFFSET(ideal, cell) < CIRCULAR_OFFSET(ideal, neighbor)) {
              *cell = *neighbor;
              cell = neighbor;
            }
          }
          return;
        }
        if (cell->key == zeroCell.key) {
          return;
        }
      }
    } else if (zeroUsed) {
      zeroUsed = false;
      population--;
    }
  }

  friend class ConstIterator;
  class ConstIterator
  {
  private:
    const HashMap& table;
    const Cell* cur;

  public:
    ConstIterator(const HashMap& tab, bool end) : table(tab), cur(0)
    {
      if (!end) {
        if (table.zeroUsed) {
          cur = &table.zeroCell;
        } else if (table.cells) {
          cur = &table.cells[-1];
          const Cell* e = table.cells + table.arraySize;
          while (++cur != e) {
            if (cur->key != table.zeroCell.key) {
              return;
            }
          }
          cur = 0;
        }
      }
    }

    inline const Cell& operator*() const { return *cur; }

    bool operator!=(const ConstIterator& it) const
    {
      return cur != it.cur;
    }

    ConstIterator& operator++()
    {
      if (cur) {
        if (cur == &table.zeroCell) {
          cur = &table.cells[-1];
        }
        const Cell* end = table.cells + table.arraySize;
        while (++cur != end) {
          if (cur->key != table.zeroCell.key) {
            return *this;
          }
        }
        cur = 0;
      }
      return *this;
    }
  };

  ConstIterator begin() const { return ConstIterator(*this, false); }
  ConstIterator end() const { return ConstIterator(*this, true); }
};

#undef FIRST_CELL
#undef CIRCULAR_NEXT
#undef CIRCULAR_OFFSET

#endif
