#pragma once

#include <vector>
#include <cassert>

namespace myhashmap
{

inline size_t Hash( int k )
{
  size_t h = *(unsigned int *)&k;     // breaks strict-aliasing
  h ^= h >> 16;
  h *= 0x85ebca6b;
  h ^= h >> 13;
  h *= 0xc2b2ae35;
  h ^= h >> 16;
  return h;
}

template< typename K, typename V >
class HashMap
{
protected:
	struct Entry
	{
		K key;
		V value;
		int next;							// index of next entry of same hash (=next collision)
	};
	std::vector< Entry > entries;	  // entries
	std::vector< int > table;				// hash table (indices to first entry)

public:
	struct Iterator
	{
		friend class HashMap;
		inline Iterator();
		template< typename U > inline bool operator ==( const U &o ) const;
		template< typename U > inline bool operator !=( const U &o ) const;
	private:
		HashMap *self;
		int index;				// entry index
	};

	// get unordered key
	inline const K &GetKey( int index ) const;

	inline bool IsEmpty() const;

	inline int GetSize() const;

  size_t GetNumCollisions() const;

	HashMap &Clear();

	HashMap &Insert( const K &key, const V &value );

	inline Iterator End();
	inline Iterator Begin();

	Iterator Find( const K &key );

protected:
	// rebuild hash
	void Rehash();
};

template< typename K, typename V >
void HashMap<K,V>::Rehash()
{
	assert( entries.empty() || !table.empty() );
	for ( size_t i=0; i<table.size(); i++ ) {
		table[i] = -1;
	}
	size_t mask = table.size()-1;
	for ( int i=0; i<(int)entries.size(); i++ ) {
		Entry &e = entries[i];
		int h = (int)(Hash(e.key) & mask);
		e.next = table[h];
		table[h] = i;
	}
}

template< typename K, typename V >
inline HashMap<K,V>::Iterator::Iterator() : self(0), index(-1) {}

template< typename K, typename V >
template< typename U >
inline bool HashMap<K,V>::Iterator::operator ==( const U &o ) const
{
	assert( self == o.self );
	return index == o.index;
}

template< typename K, typename V >
template< typename U >
inline bool HashMap<K,V>::Iterator::operator !=( const U &o ) const
{
	assert( self == o.self );
	return index != o.index;
}

template< typename K, typename V >
inline const K &HashMap<K,V>::GetKey( int index ) const
{
	return entries[index].key;
}

template< typename K, typename V >
inline bool HashMap<K,V>::IsEmpty() const
{
	return entries.IsEmpty();
}

template< typename K, typename V >
inline int HashMap<K,V>::GetSize() const
{
	return (int)entries.size();
}

template< typename K, typename V >
HashMap<K,V> &HashMap<K,V>::Clear()
{
	entries.clear();
	table.clear();
}

template< typename K, typename V >
HashMap<K,V> &HashMap<K,V>::Insert( const K &key, const V &value )
{
	size_t h = Hash(key);
	int sz = (int)table.size();
	Entry e;
	e.key = key;
	e.value = value;
	e.next = -1;
	if ( !sz ) {
		// just add one entry (empty table)
		entries.push_back( e );
		table.push_back( 0 );
		return *this;
	}
	size_t index = h & ((size_t)sz-1);
	int tmp = table[(int)index];
	while ( tmp >= 0 ) {
		Entry &he = entries[tmp];
		if ( he.key == key ) {
			// already in table
			he.value = value;
			return *this;
		}
		tmp = he.next;
	}
	e.next = table[(int)index];
	table[index] = entries.size();
	entries.push_back(e);
	// rehash if too big
	if ( entries.size() >= table.size() ) {
		assert( sz*2 > sz );
		table.resize( sz*2 );
		Rehash();
	}
	assert( Find(key) != End() );
	return *this;
}

template< typename K, typename V >
inline typename HashMap<K,V>::Iterator HashMap<K,V>::End()
{
	Iterator res;
	res.self = this;
	res.index = -1;
	return res;
}

template< typename K, typename V >
inline typename HashMap<K,V>::Iterator HashMap<K,V>::Begin()
{
	assert( !IsEmpty() );
	Iterator res;
	res.self = this;
	res.index = 0;
	return res;
}

template< typename K, typename V >
size_t HashMap<K,V>::GetNumCollisions() const
{
	size_t res = 0;
	// get total number of collisions
	for ( int i=0; i<entries.size(); i++ ) {
		res += entries[i].next >= 0;
	}
	return res;
}

template< typename K, typename V >
typename HashMap<K,V>::Iterator HashMap<K,V>::Find( const K &key )
{
	int sz = (int)table.size();
	if ( !sz ) {
		return End();
	}
	size_t h = Hash(key) & ((size_t)sz-1);
	int index = table[(int)h];
	while ( index >= 0 ) {
		const Entry &e = entries[index];
		if ( e.key == key ) {
			Iterator res;
			res.self = this;
			res.index = index;
			return res;
		}
		index = e.next;
	}
	return End();
}

}
